/*
 *  Linux kernel adaptation layers for owl wifi library.
 *  Copyright (C) 2010 HD Wireless AB
 *
 */

#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/version.h>
#include <net/iw_handler.h>

#include "owl_sync.h"
#include "owl_wext_droid.h"

/* MHz */
#define channel_to_freq(c) (2412 + ((c - 1) * 5))

int owl_wext_get_name(struct net_device *dev, struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        strncpy(wrqu->name, "IEEE 802.11bg", IFNAMSIZ);
        return 0;
}


int owl_wext_get_range(struct net_device *dev, struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
        struct iw_range *range = (struct iw_range *) extra;
        int i;
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        range->we_version_compiled = WIRELESS_EXT;
        range->we_version_source = 22;

        range->num_channels = 14;
        range->num_frequency = 14;
        for (i = 0; i < range->num_channels; i++) {
                range->freq[i].m = channel_to_freq(i + 1);
                range->freq[i].e = 6;
                range->freq[i].i = i + 1;
        }
                
        range->max_qual.qual = 30; /* max snr */
        range->avg_qual.qual = 15; /* avg snr */

        range->num_encoding_sizes = 32;
        range->max_encoding_tokens = 4;

        /* used by wpa_supplicant */
        range->enc_capa =
                IW_ENC_CAPA_WPA |
                IW_ENC_CAPA_WPA2 |
                IW_ENC_CAPA_CIPHER_TKIP	|
                IW_ENC_CAPA_CIPHER_CCMP;
        
        return 0;
}

int owl_wext_set_scan(struct net_device *dev, struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        wl_err_t wl_err;
        struct owl_net *priv = netdev_priv(dev);
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        
        wl_err = owl_sync_wl_scan(priv);

        /* This will only happen if invoked while when wl_connect() is trying to
         * connect to a net that is not currently in the scan list; forcing a
         * scan to look for it. Otherwise, wl_scan() is only invoked through
         * owl_sync_wl_scan(). We can safely return 0 here, allowing e.g.
         * iwconfig to fetch the list. Note that any ongoing scan, initiated
         * by wl_connect() will not invoke the owl_sync:event_cb().
         */
        if (wl_err == WL_RETRY)
                return 0;

        /* case where scanning is initiated when interface is down */
        else if (wl_err != WL_SUCCESS) {
                dev_info(&dev->dev, "scan failed, err=%d\n", wl_err);
                return -EIO;
        }

        /* owl_sync:event_cb() will make sure that the SIOCGIWSCAN event
         * is signalled
         */
        
        return 0;
}


int owl_wext_get_essid(struct net_device *dev, struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
        struct wl_network_t *net;
        struct owl_net *priv = netdev_priv(dev);
        int err;
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        /* just set any bit that is not in the IW_ENCODE_INDEX range */
        wrqu->essid.flags = IW_ENCODE_FLAGS;

        if ((err = owl_sync_wl_get_current_network(priv, &net)) < 0)
                return err;
                
        if (net == NULL) {
                wrqu->essid.length = 0;
                return 0;
        }
                
        wrqu->essid.length = net->ssid.len;
        memcpy(extra, net->ssid.ssid, net->ssid.len);
        kfree(net);
        return 0;
}


/*
 * owl_sync_wl_connect() will disconnect from the current AP before trying
 * to connect to the provided SSID (even though the SSID is the same as the
 * one we're connected to.
 *
 */
int owl_wext_set_essid(struct net_device *dev, struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
        char *ssid = extra;
        wl_err_t wl_err;
        struct owl_net *priv = netdev_priv(dev);
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        if (wrqu->essid.length == 0) {
                dev_dbg(&dev->dev, "disconnecting from current net\n");

                /* safe to call when not connected as well */
                wl_err = owl_sync_wl_disconnect(priv);
        }
        else {
                dev_dbg(&dev->dev, "connecting\n");
                wl_err = owl_sync_wl_connect(priv, ssid, wrqu->essid.length);
        }

        switch (wl_err) {
        case WL_SUCCESS:
        case WL_NOEFFECT:
                return 0;

        case WL_RETRY:
                /* connect in progress */
                dev_info(&dev->dev, "busy, connect in progress\n");
                return -EBUSY;

        case WL_FAILURE:
                /* WL_FAILURE could be e.g incompatible authentication mode,
                 * and should be ignored
                 */
                dev_info(&dev->dev, "connect failed\n");
                return 0;

        default:
                dev_warn(&dev->dev, "unhandled error, err=%d\n", wl_err);
                return -EIO;

        };
}


int owl_wext_set_mode(struct net_device *dev, struct iw_request_info *info,
                    union iwreq_data *wrqu, char *extra)
{
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        switch (wrqu->mode) {
        case IW_MODE_INFRA:
        case IW_MODE_AUTO:
                return 0;
        default:
                return -EINVAL;
        }
}


int owl_wext_get_mode(struct net_device *dev, struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        wrqu->mode = IW_MODE_INFRA;
        return 0;
}

/*
 * owl_sync_wl_connect_bssid() will disconnect from the current AP before trying
 * to connect to the provided BSSID (even though the BSSID is the same as the
 * one we're connected to.
 *
 */
int owl_wext_set_ap(struct net_device *dev, struct iw_request_info *info,
                    union iwreq_data *wrqu, char *extra)
{
        struct wl_mac_addr_t bssid;
        struct owl_net *priv = netdev_priv(dev);
                
        dev_dbg(&dev->dev, "ENTER %s bssid: %02x %02x %02x %02x %02x %02x\n",
                __func__,
                wrqu->ap_addr.sa_data[0],
                wrqu->ap_addr.sa_data[1],
                wrqu->ap_addr.sa_data[2],
                wrqu->ap_addr.sa_data[3],
                wrqu->ap_addr.sa_data[4],
                wrqu->ap_addr.sa_data[5]);

        /* wpa_supplicant driver_wext:wpa_driver_wext_associate()
         * explicitly clears the BSSID when about to connect
         * to an SSID. This is not necessary for wl_api since when
         * invoking wl_connect() and wl_connect_bssid(), both the
         * desired ssid and bssid are updated. Further, wpa_supplicant
         * want's this request to be successful so we better just return 0
         * here. If we would pass through the call to '
         * owl_sync_wl_connect_bssid(), we would get an error back since
         * there's already a connection in progress for an SSID, issued by
         * wpa_supplicant.
         *
         * Note that this is also the way which wpa_supplicant >= 0.6.10
         * requests a disconnect.
         */
        if (is_zero_ether_addr(wrqu->ap_addr.sa_data)) {
                dev_dbg(&dev->dev, "NULL bssid - disconn (if conn)\n");
                (void) owl_sync_wl_disconnect(priv);
                return 0;
        }
        
        memcpy(bssid.octet, wrqu->ap_addr.sa_data, ETH_ALEN);

        /* owl_sync_wl_connect_bssid() will return an error if we're currently
         * connecting (to an ssid or bssid). wpa_supplicant will invoke
         * both set_ssid and set_ap so we'll not propagate any error codes
         * to please the supplicant.
         *
         * To avoid problems when wpa_supplicant connects throught set_ssid
         * and then set_ap, owl_sync_wl_connect_bssid() will _not_ reconnect
         * to the current network. This will avoid a disconnect and key clearing
         * if we manage to finish the connection issued through set_ssid at the
         * time set_ap is called.
         */
        (void) owl_sync_wl_connect_bssid(priv, bssid);
        return 0;
}

int owl_wext_get_ap(struct net_device *dev, struct iw_request_info *info,
                    union iwreq_data *wrqu, char *extra)
{
        struct wl_network_t *net;
        int err;
        struct owl_net *priv = netdev_priv(dev);

        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        wrqu->ap_addr.sa_family = ARPHRD_ETHER;
        if ((err = owl_sync_wl_get_current_network(priv, &net)) < 0)
                return err;

        /* no network */
        if (net == NULL) {
                memset(wrqu->ap_addr.sa_data, 0, ETH_ALEN);
        } else {
                memcpy(wrqu->ap_addr.sa_data, net->bssid.octet, ETH_ALEN);
                kfree(net);
        }

        return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
# define IWE_STREAM_ADD_POINT(a, b, c, d, e) iwe_stream_add_point(a, b, c, d, e)
# define IWE_STREAM_ADD_EVENT(a, b, c, d, e) iwe_stream_add_event(a, b, c, d, e)
#else
# define IWE_STREAM_ADD_POINT(a, b, c, d, e) iwe_stream_add_point(b, c, d, e)
# define IWE_STREAM_ADD_EVENT(a, b, c, d, e) iwe_stream_add_event(b, c, d, e)
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27) */

int owl_wext_get_scan(struct net_device *dev, struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        struct wl_network_list_t netlist;
        int i;
        int err = 0;

        /* for IWEVCUSTOM */
        char custom[64];
        int len;

        struct iw_event iwe;
        char *p;
        char *stream = extra;
        char *ends = extra + wrqu->data.length;
        struct owl_net *priv = netdev_priv(dev);
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        if ((err = owl_sync_wl_get_network_list(priv, &netlist)))
                return -EIO;
        
        for (i = 0; i < netlist.cnt; i++) {
                struct wl_network_t *net = netlist.net[i];

                iwe.cmd = SIOCGIWAP;
                iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
                memcpy(iwe.u.ap_addr.sa_data, net->bssid.octet, ETH_ALEN);
                p = IWE_STREAM_ADD_EVENT(info, stream, ends, &iwe,
                                         IW_EV_ADDR_LEN);
                if (p == stream) {
                        err = -E2BIG;
                        goto out;
                }
                stream = p;

                iwe.cmd = SIOCGIWESSID;
                iwe.u.essid.length = net->ssid.len;
                /* just set any bit that is not in the IW_ENCODE_INDEX range */
                iwe.u.essid.flags = IW_ENCODE_FLAGS;
                p = IWE_STREAM_ADD_POINT(info, stream, ends, &iwe,
                                         net->ssid.ssid);
                if (p == stream) {
                        err = -E2BIG;
                        goto out;
                }
                stream = p;

                iwe.cmd = SIOCGIWMODE;
                iwe.u.mode = IW_MODE_INFRA; /* only BSS in netlist */
                p = IWE_STREAM_ADD_EVENT(info, stream, ends, &iwe,
                                         IW_EV_UINT_LEN);
                if (p == stream) {
                        err = -E2BIG;
                        goto out;
                }
                stream = p;

                iwe.cmd = SIOCGIWENCODE;
                if (net->enc_type != ENC_TYPE_NONE)
                        iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
                else
                        iwe.u.data.flags = IW_ENCODE_DISABLED;
                iwe.u.data.length = 0;
                p = IWE_STREAM_ADD_POINT(info, stream, ends, &iwe, "");
                if (p == stream) {
                        err = -E2BIG;
                        goto out;
                }
                stream = p;

                iwe.cmd = SIOCGIWFREQ;
                iwe.u.freq.m = channel_to_freq(net->channel);
                iwe.u.freq.e = 6;
                iwe.u.freq.i = net->channel;
                p = IWE_STREAM_ADD_EVENT(info, stream, ends, &iwe,
                                         IW_EV_FREQ_LEN);
                if (p == stream) {
                        err = -E2BIG;
                        goto out;
                }
                stream = p;
                
                iwe.cmd = IWEVQUAL;
                iwe.u.qual.qual = net->snr != WL_SNR_UNKNOWN ? net->snr : 0;
                iwe.u.qual.level = net->rssi;
                iwe.u.qual.noise =
                    net->snr != WL_SNR_UNKNOWN ? net->rssi - net->snr : 0;
                p = IWE_STREAM_ADD_EVENT(info, stream, ends, &iwe,
                                         IW_EV_QUAL_LEN);
                if (p == stream) {
                        err = -E2BIG;
                        goto out;
                }
                stream = p;

                if (net->ie_len && net->ie_len <= IW_GENERIC_IE_MAX) {
                        iwe.cmd = IWEVGENIE;
                        iwe.u.data.flags = 0;
                        iwe.u.data.length = net->ie_len;
                        p = IWE_STREAM_ADD_POINT(info, stream, ends, &iwe,
                                                 net->ie);
                        if (p == stream) {
                                err = -E2BIG;
                                goto out;
                        }
                        stream = p;
                }
                
                iwe.cmd = IWEVCUSTOM;
                len = snprintf(custom, sizeof(custom),
                               "Beacon period: %d Kusec", net->beacon_period);
                iwe.u.data.length = len;
                iwe.u.data.flags = 0;
                p = IWE_STREAM_ADD_POINT(info, stream, ends, &iwe, custom);
                if (p == stream) {
                        err = -E2BIG;
                        goto out;
                }
                stream = p;

                if (net->dtim_period) {
                        iwe.cmd = IWEVCUSTOM;
                        len = snprintf(custom, sizeof(custom),
                                       "DTIM period: %d", net->dtim_period);
                        iwe.u.data.length = len;
                        iwe.u.data.flags = 0;
                        p = IWE_STREAM_ADD_POINT(info, stream, ends, &iwe,
                                                 custom);
                        if (p == stream) {
                                err = -E2BIG;
                                goto out;
                        }
                        stream = p;
                }
        }

out:
        owl_sync_put_network_list(&netlist);
        return err;
}

int owl_wext_set_freq(struct net_device *dev,
                      struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        /* wl_api will use frequency required by net anyway */
        return 0;
}

int owl_wext_get_freq(struct net_device *dev,
                      struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        struct wl_network_t *net;
        int err;
        struct owl_net *priv = netdev_priv(dev);
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        if ((err = owl_sync_wl_get_current_network(priv, &net)) < 0)
                return err;

        if (net == NULL) /* no network */
                return -EIO;

        wrqu->freq.m = channel_to_freq(net->channel);
        wrqu->freq.e = 6;
        wrqu->freq.i = net->channel;
        kfree(net);
        return 0;
}


int owl_wext_set_power(struct net_device *dev,
                       struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
        wl_err_t wl_err;
        struct owl_net *priv = netdev_priv(dev);

        dev_dbg(&dev->dev, "%s flags: %x disabled: %d value: %d\n",
                __func__, wrqu->power.flags, wrqu->power.disabled,
                wrqu->power.value);

        /* iwconfig doesn't init the struct properly when disabled is set */
        if (wrqu->power.disabled) {
                wrqu->power.flags = 0;
                wrqu->power.value = 0;
        }
        
        /* change traffic timeout? */
        if (wrqu->power.flags & IW_POWER_TIMEOUT) {
                unsigned int tmo = wrqu->power.value;
                dev_dbg(&dev->dev, "setting traffic timeout to %d\n", tmo);
                
                /* everything except traffic timeout are default values */
                if ((wl_err = owl_sync_wl_conf_ps(priv, 0, tmo, 5000, 1, 20)) !=
                    WL_SUCCESS) {
                        dev_info(&dev->dev, "failed to set ps traffic timeout "
                                 "(%d), err=%d\n", tmo, wl_err);
                        return -EIO;
                }
                priv->ps_traffic_timeout = tmo;
        }

        /* already disabled? */
        if (wrqu->power.disabled && !priv->ps_enabled)
                return 0;

        /* power save already enabled? */
        if (!wrqu->power.disabled && priv->ps_enabled)
                return 0;

        /* disable power save? */
        if (wrqu->power.disabled) {
                if ((wl_err = owl_sync_wl_disable_ps(priv)) != WL_SUCCESS) {
                        dev_err(&dev->dev, "failed to disable ps, err=%d\n",
                                wl_err);
                        return -EIO;
                }
                priv->ps_enabled = 0;
                return 0;
        }

        /* all other cases tested, then enable power save */
        if ((wl_err = owl_sync_wl_enable_ps(priv)) != WL_SUCCESS) {
                dev_err(&dev->dev, "failed to enable ps, err=%d\n",
                        wl_err);
                return -EIO;
        }
        priv->ps_enabled = 1;
        return 0;
}


int owl_wext_get_power(struct net_device *dev,
                       struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
        struct owl_net *priv = netdev_priv(dev);

        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        /* use IW_POWER_RELATIVE to avoid issues with getting iwconfig to
         * print the correct unit
         */
        wrqu->power.flags = IW_POWER_TIMEOUT | IW_POWER_RELATIVE;
        wrqu->power.value = priv->ps_traffic_timeout;
        wrqu->power.disabled = !priv->ps_enabled;
        return 0;        
}


int owl_wext_get_rate(struct net_device *dev,
                      struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        wl_rate_t tx_rate;
        struct owl_net *priv = netdev_priv(dev);
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        
        if (owl_sync_wl_get_rate(priv, &tx_rate, NULL) != WL_SUCCESS)
                return -EIO;

        if (tx_rate == WL_RATE_INVALID)
                return -EIO;

        wrqu->bitrate.value = tx_rate * 500000;
        return 0;        
}

static int set_auth_mode(struct owl_net *priv)
{
        enum wl_auth_mode auth_mode = AUTH_MODE_OPEN_SYSTEM;
        
        /* Note that param IW_AUTH_WPA_ENABLED is just used by wpa_supplicant
         * for backwards compatibility. This is called only just after init
         * and just before deinit. However, IW_AUTH_WPA_VERSION will be updated
         * upon each association. Therefore, we'll use IW_AUTH_WPA_VERSION
         * to determine whether WPA is used or not.
         */
        
        if (priv->auth_param[IW_AUTH_WPA_VERSION] == IW_AUTH_WPA_VERSION_WPA ||
            priv->auth_param[IW_AUTH_WPA_VERSION] == IW_AUTH_WPA_VERSION_WPA2) {
                /* let wl_api figure out the authentication mode to use */
                auth_mode = AUTH_MODE_AUTO;
                
        } else {
                /* non-WPA */
                if (priv->auth_param[IW_AUTH_80211_AUTH_ALG] ==
                    IW_AUTH_ALG_OPEN_SYSTEM)
                    auth_mode = AUTH_MODE_OPEN_SYSTEM;
                else if (priv->auth_param[IW_AUTH_80211_AUTH_ALG] ==
                         IW_AUTH_ALG_SHARED_KEY)
                    auth_mode = AUTH_MODE_SHARED_KEY;
        }

        /* just to be able to report back to iwconfig */
        if (auth_mode == AUTH_MODE_SHARED_KEY)
                priv->auth_mode = AUTH_MODE_SHARED_KEY;
        else
                priv->auth_mode = AUTH_MODE_OPEN_SYSTEM;

        return 0;
}


/*
 * iwconfig owl0 key on
 * iwconfig owl0 key [off] [open] [restricted] [index] [key]
 *
 * As of 1.02, the wl_api library will enfore shared key authentication
 * to be used with wep. Hence it is no longer possible to use WEP encryption
 * with open authentication. The comments below are kept just for historical
 * reasons. 'restricted' will therefore not have any effect.
 * 
 * Note that if no keys were configured at the time when the STA connected
 * to the AP, it is necessary to re-associate with the AP for the newly
 * added keys to take effect. However, if a key was configured prior to
 * association (even though the key was not correct), any newly added key
 * should be used without requiring a re-association. This scenario can occur
 * if WEP is used in open authentication mode. For shared key authentication,
 * the association will fail if the key is not configured (or incorrect).
 *
 * So, in the case we have an AP with open authentication that requires the
 * key 0123456789 [1], the following scenario will leave the STA associated,
 * but unable to transfer data frames:
 * 'iwconfig owl0 essid ap; iwconfig owl0 key 0123456789'
 * To recover, a re-association is needed:
 * 'iwconfig owl0 essid ap'
 * A working sequence is e.g:
 * 'iwconfig owl0 key 0123456789; iwconfig owl0 essid ap'
 * Another working sequence is e.g:
 * 'iwconfig owl0 key 0000011111; iwconfig owl0 essid ap;
 *  iwconfig owl0 key 0123456789'
 *
 *
 * For shared key authentication, the STA transmit key index will decide the
 * key to be used for authentication. The AP must have a matching key configured
 * at the same slot index for authentication to succeed.
 * 
 * In general, when data frames are transmitted, the transmit (default) key is
 * used to encode the data. The index of the transmit key can be different on
 * the STA and the AP. The STA/AP must have a matching key configured at the
 * slot index used by the other ends transmit key in order to decrypt the data.
 *
 * If the AP is configured with open system authentication, association should
 * be successful even though the STA is configured to use shared key
 * authentication, that is 'iwconfig owl0 key restricted'.
 * When the AP is configured to use shared key authentication,
 * the STA must also use shared key authentication and have the proper key set
 * in order to assicate successfully.
 * 
 */
int owl_wext_set_encode(struct net_device *dev,
                        struct iw_request_info *info,
                        union iwreq_data *wrqu, char *extra)
{
        unsigned char *key = extra;
        int len = wrqu->encoding.length;
        unsigned int idx;
        struct wl_mac_addr_t bssid;
        struct owl_net *priv = netdev_priv(dev);
        
        dev_dbg(&dev->dev, "%s length: %d idx: %d flags: %s%s%s%s\n",
                __func__, len, wrqu->encoding.flags & IW_ENCODE_INDEX,
                wrqu->encoding.flags & IW_ENCODE_DISABLED ?
                "IW_ENCODE_DISABLED " : "",
                wrqu->encoding.flags & IW_ENCODE_OPEN ?
                "IW_ENCODE_OPEN" : "",
                wrqu->encoding.flags & IW_ENCODE_RESTRICTED ?
                "IW_ENCODE_RESTRICTED" : "",
                wrqu->encoding.flags & IW_ENCODE_NOKEY ?
                "IW_ENCODE_NOKEY" : "");

#ifdef DEBUG
        print_hex_dump_bytes("key ", DUMP_PREFIX_NONE, key, len);
#endif
        
        /* wl_api valid key indices are 0-3, but iwconfig uses 1-4,
         * if no idx specified, default to 0 when adding key
         */
        idx = wrqu->encoding.flags & IW_ENCODE_INDEX;
        if (idx)
                idx--;
        
        if (idx > 3) {
                dev_warn(&dev->dev, "invalid key idx\n");
                return -EINVAL;
        }
        
        if (len > OWL_MAX_KEY_LEN) {
                dev_warn(&dev->dev, "invalid key len\n");
                return -EINVAL;
        }
                        
        /* can't add key-mapping keys through iwconfig */
        memset(bssid.octet, 0xff, sizeof(bssid));

        /* disable enc, then delete all keys */
        if (wrqu->encoding.flags & IW_ENCODE_DISABLED) {
                int i;
                for (i = 0; i < 4; i++) {
                        if (owl_sync_wl_delete_wep_key(priv, i, &bssid) !=
                            WL_SUCCESS) {
                                dev_info(&dev->dev, "failed to delete wep key "
                                         "%d\n", i);
                                return -EIO;
                        }
                        priv->key_info[i].len = 0;
                }
                
                priv->default_key = 0;
        }


        /* set shared key authentication */
        if (wrqu->encoding.flags & IW_ENCODE_RESTRICTED)
                priv->auth_param[IW_AUTH_80211_AUTH_ALG] =
                        IW_AUTH_ALG_SHARED_KEY;

        /* or set open system authentication (no effect if AP uses WEP) */
        else if (wrqu->encoding.flags & IW_ENCODE_OPEN)
                priv->auth_param[IW_AUTH_80211_AUTH_ALG] =
                        IW_AUTH_ALG_OPEN_SYSTEM;
        
        /* Has key, then add it.
         *
         * Note that the wpa_supplicant function
         * driver_wext:wpa_driver_wext_auth_alg_fallback() might invoke this
         * function with flags and key set to 0, in this case we're only
         * supposed to set the authentication mode, not add any keys.
         */
        if (!(wrqu->encoding.flags & IW_ENCODE_NOKEY) && len) {
                if (owl_sync_wl_add_wep_key(priv, idx, len, key, &bssid) !=
                    WL_SUCCESS) {
                        dev_info(&dev->dev, "failed to add wep key "
                                 "idx=%d len=%d\n", idx, len);
                        return -EINVAL;
                }

                memcpy(priv->key_info[idx].key, key, len);
                priv->key_info[idx].len = len;
                
                /* adding a non-key-mapping key implicitly sets the default
                 * key as well
                 */
                priv->default_key = idx;
        }

        /* no key, but an index is specified */
        else if (wrqu->encoding.flags & IW_ENCODE_INDEX) {
                if (priv->key_info[idx].len == 0) {
                        dev_info(&dev->dev, "key %d is not configured\n", idx);
                        return -EINVAL;
                }

                if (owl_sync_wl_set_default_wep_key(priv, idx) != WL_SUCCESS) {
                        dev_info(&dev->dev, "failed to set default key "
                                 "idx=%d\n", idx);
                        return -EINVAL;
                }
                priv->default_key = idx;
        }

        return set_auth_mode(priv);
}


int owl_wext_get_encode(struct net_device *dev,
                        struct iw_request_info *info,
                        union iwreq_data *wrqu, char *extra)
{
        char *key = extra;
        unsigned int idx;
        struct owl_net *priv = netdev_priv(dev);

        dev_dbg(&dev->dev, "ENTER %s idx=%d\n", __func__,
                wrqu->encoding.flags & IW_ENCODE_INDEX);

        /* iwlist keys sets the the index of the key that should be
         * displayed
         */
        idx = wrqu->encoding.flags & IW_ENCODE_INDEX;
        if (idx)
                idx--;
        else
                idx = priv->default_key;

        if (idx > 3) {
                dev_warn(&dev->dev, "invalid key index\n");
                return -EINVAL;
        }

        if (priv->auth_mode == AUTH_MODE_SHARED_KEY)
                wrqu->encoding.flags |= IW_ENCODE_RESTRICTED;
        else
                wrqu->encoding.flags |= IW_ENCODE_OPEN;
        
        wrqu->encoding.flags |= (idx + 1);

        if (wrqu->encoding.length >= priv->key_info[idx].len) {
                memcpy(key, priv->key_info[idx].key, priv->key_info[idx].len);
                wrqu->encoding.length = priv->key_info[idx].len;
        } else {
                dev_warn(&dev->dev, "failed to copy key\n");
        }
        
        return 0;
}

int owl_wext_set_encode_ext(struct net_device *dev,
                            struct iw_request_info *info,
                            union iwreq_data *wrqu, char *extra)
{
        struct iw_encode_ext *ext = (struct iw_encode_ext *) extra;
        struct owl_net *priv = netdev_priv(dev);
        struct wl_mac_addr_t bssid;
        struct wl_receive_seq_cnt_t rsc, *rscp = NULL;
        int tx_key = 0;
        int idx;

        dev_dbg(&dev->dev, "%s key_len: %d idx: %d ext_flags: %s%s%s%s\n",
                __func__,
                ext->key_len,
                wrqu->encoding.flags & IW_ENCODE_INDEX,
                ext->ext_flags & IW_ENCODE_EXT_TX_SEQ_VALID ?
                "IW_ENCODE_EXT_TX_SEQ_VALID " : "",
                ext->ext_flags & IW_ENCODE_EXT_RX_SEQ_VALID ?
                "IW_ENCODE_EXT_RX_SEQ_VALID " : "",
                ext->ext_flags & IW_ENCODE_EXT_GROUP_KEY ?
                "IW_ENCODE_EXT_GROUP_KEY " : "",
                ext->ext_flags & IW_ENCODE_EXT_SET_TX_KEY ?
                "IW_ENCODE_EXT_SET_TX_KEY " : "");

#ifdef DEBUG
        print_hex_dump_bytes("key: ", DUMP_PREFIX_NONE, ext->key, ext->key_len);
#endif

        /* wl_api valid key indices are 0-3, but wext uses 1-4,
         * if no idx specified, default to 0 when adding key
         */
        idx = wrqu->encoding.flags & IW_ENCODE_INDEX;
        if(idx)
                idx--;

        if (idx > 3) {
                dev_warn(&dev->dev, "invalid key index\n");
                return -EINVAL;
        }
        
        if (ext->key_len > OWL_MAX_KEY_LEN) {
                dev_warn(&dev->dev, "invalid key len: %d\n", ext->key_len);
                return -EINVAL;
        }        

        if (ext->ext_flags & IW_ENCODE_EXT_SET_TX_KEY)
                tx_key = 1;
        
        if (ext->ext_flags & IW_ENCODE_EXT_RX_SEQ_VALID) {
                memcpy(rsc.octet, ext->rx_seq, IW_ENCODE_SEQ_MAX_SIZE);
                rscp = &rsc;
        }

        memcpy(bssid.octet, ext->addr.sa_data, ETH_ALEN);

        /* will remove key if len is 0 */
        if (owl_sync_wl_set_key(priv, idx, ext->key_len, ext->key, 1, &bssid,
                                rscp, tx_key) != WL_SUCCESS) {
                dev_err(&dev->dev, "failed to set key\n");
                return -EINVAL;
        }
        
        memcpy(priv->key_info[idx].key, ext->key, ext->key_len);
        priv->key_info[idx].len = ext->key_len;

        return set_auth_mode(priv);
}

int owl_wext_set_genie(struct net_device *dev,
                       struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
#ifdef DEBUG
        char *genie = extra;
        int len = wrqu->data.length;

        print_hex_dump_bytes("genie ", DUMP_PREFIX_NONE, genie, len);
#endif
        return 0;
}


int owl_wext_set_auth(struct net_device *dev,
                       struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
        int index = wrqu->param.flags & IW_AUTH_INDEX;
        int value = wrqu->param.value;
        struct owl_net *priv = netdev_priv(dev);
        
#ifdef DEBUG
        int flags = wrqu->param.flags & IW_AUTH_FLAGS;

        const char *iw_auth_param[] = {
                "WPA_VERSION",
                "CIPHER_PAIRWISE",
                "CIPHER_GROUP",
                "KEY_MGMT",
		"TKIP_COUNTERMEASURES",
                "DROP_UNENCRYPTED",
                "80211_AUTH_ALG",
		"WPA_ENABLED",
                "RX_UNENCRYPTED_EAPOL",
                "ROAMING_CONTROL",
                "PRIVACY_INVOKED",
                "CIPHER_GROUP_MGMT",
                "MFP"
        };

        dev_dbg(&dev->dev, "%s index=%03x flags=%01x value=%08x %s:\n",
                __func__, index, flags, value,
                index < ARRAY_SIZE(iw_auth_param) ?
                iw_auth_param[index] : "unknown");

        switch (index) {
        case IW_AUTH_WPA_VERSION:
                dev_dbg(&dev->dev, "%sDISABLED | %sWPA | %sWPA2\n",
                        value & IW_AUTH_WPA_VERSION_DISABLED ? "" : "~",
                        value & IW_AUTH_WPA_VERSION_WPA ? "" : "~",
                        value & IW_AUTH_WPA_VERSION_WPA2 ? "" : "~");
                break;

        case IW_AUTH_CIPHER_PAIRWISE:
        case IW_AUTH_CIPHER_GROUP:
        case IW_AUTH_CIPHER_GROUP_MGMT:
                dev_dbg(&dev->dev, "%sNONE | %sWEP40 | %sTKIP | %sCCMP | "
                        "%sWEP104 | %sAES_CMAC\n",
                        value & IW_AUTH_CIPHER_NONE ? "" : "~",
                        value & IW_AUTH_CIPHER_WEP40 ? "" : "~",
                        value & IW_AUTH_CIPHER_TKIP ? "" : "~",
                        value & IW_AUTH_CIPHER_CCMP ? "" : "~",
                        value & IW_AUTH_CIPHER_WEP104 ? "" : "~",
                        value & IW_AUTH_CIPHER_AES_CMAC ? "" : "~");
                break;
                
        case IW_AUTH_80211_AUTH_ALG:
                dev_dbg(&dev->dev, "%sOPEN_SYSTEM | %sSHARED_KEY | %sLEAP\n",
                        value & IW_AUTH_ALG_OPEN_SYSTEM ? "" : "~",
                        value & IW_AUTH_ALG_SHARED_KEY ? "" : "~",
                        value & IW_AUTH_ALG_LEAP ? "" : "~");
                break;

        case IW_AUTH_KEY_MGMT:
                dev_dbg(&dev->dev, "%s802_1X | %sPSK\n",
                        value & IW_AUTH_KEY_MGMT_802_1X ? "" : "~",
                        value & IW_AUTH_KEY_MGMT_PSK ? "" : "~");
                break;
                
        case IW_AUTH_DROP_UNENCRYPTED:
        case IW_AUTH_PRIVACY_INVOKED:
        case IW_AUTH_RX_UNENCRYPTED_EAPOL:
        case IW_AUTH_TKIP_COUNTERMEASURES:
        case IW_AUTH_WPA_ENABLED:
                dev_dbg(&dev->dev, "%d\n", value);
                break;
                
        default:
                dev_dbg(&dev->dev, "no trace available for param %d\n", index);
                
        }
#endif

        if (index >= ARRAY_SIZE(priv->auth_param)) {
                dev_warn(&dev->dev, "auth param out of range\n");
                return -EINVAL;
        }

        priv->auth_param[index] = value;

        return set_auth_mode(priv);
}

/* wpa_supplicant 0.6.7 relies on set_mlme() for disconnecting.
 * wpa_supplicant 0.6.10 is able to fall back on set_ap() and set_essid().
 *
 */
int owl_wext_set_mlme(struct net_device *dev,
                      struct iw_request_info *info,
                      union iwreq_data *wrqu, char *extra)
{
        struct iw_mlme *mlme = (struct iw_mlme *) extra;
        struct owl_net *priv = netdev_priv(dev);

#ifdef DEBUG
        const char *mlme_cmd[] = {
                "IW_MLME_DEAUTH",
                "IW_MLME_DISASSOC",
                "IW_MLME_AUTH",
                "IW_MLME_ASSOC"
        };
        
        dev_dbg(&dev->dev, "%s cmd: %s reason_code: %d\n", __func__,
                mlme->cmd < ARRAY_SIZE(mlme_cmd) ?
                mlme_cmd[mlme->cmd] : "unknown",
                mlme->reason_code);
#endif

        switch (mlme->cmd) {
        case IW_MLME_DEAUTH:
        case IW_MLME_DISASSOC:
                /* will disconnect and wait for completion */
                (void) owl_sync_wl_connect(priv, NULL, 0);
                break;

        case IW_MLME_AUTH:
        case IW_MLME_ASSOC:
        default:
                dev_info(&dev->dev, "mlme cmd: %d not handled\n", mlme->cmd);
                return -EINVAL;
        }

        return 0;
}

struct iw_statistics *owl_wext_get_wireless_stats(struct net_device *dev)
{
        struct owl_net *priv = netdev_priv(dev);
        struct iw_statistics *stats = &priv->wireless_stats;
        struct wl_network_t *net;
        int err;
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        if ((err = owl_sync_wl_get_current_network(priv, &net)) < 0)
                return NULL;

        if (net == NULL) /* no network */
                return NULL;

        stats->qual.updated = IW_QUAL_DBM | IW_QUAL_QUAL_UPDATED |
                IW_QUAL_LEVEL_UPDATED | IW_QUAL_NOISE_UPDATED;
        stats->qual.qual = net->snr != WL_SNR_UNKNOWN ? net->snr : 0;
        stats->qual.level = net->rssi;
        stats->qual.noise = 
                net->snr != WL_SNR_UNKNOWN ? net->rssi - net->snr : 0;

        kfree(net);
        return stats;
}


const iw_handler owl_iw_handler[] = {
        NULL,                           /* 0x8b00 SIOCSIWCOMMIT */
        owl_wext_get_name,              /* 0x8b01 SIOCGIWNAME   */

        NULL,                           /* 0x8b02 SIOCSIWNWID   */
        NULL,                           /* 0x8b03 SIOCGIWNWID   */
        owl_wext_set_freq,              /* 0x8b04 SIOCSIWFREQ   */
        owl_wext_get_freq,              /* 0x8b05 SIOCGIWFREQ   */
        owl_wext_set_mode,              /* 0x8b06 SIOCSIWMODE   */
        owl_wext_get_mode,              /* 0x8b07 SIOCGIWMODE   */
        NULL,                           /* 0x8b08 SIOCSIWSENS   */
        NULL,                           /* 0x8b09 SIOCGIWSENS   */

        NULL,                           /* 0x8b0a SIOCSIWRANGE  */
        owl_wext_get_range,             /* 0x8b0b SIOCGIWRANGE  */
        owl_wext_set_droid,             /* 0x8b0c SIOCSIWPRIV   */
        NULL,                           /* 0x8b0d SIOCGIWPRIV   */
        NULL,                           /* 0x8b0e SIOCSIWSTATS  */
        NULL,                           /* 0x8b0f SIOCGIWSTATS  */

        NULL,                           /* 0x8b10 SIOCSIWSPY    */
        NULL,                           /* 0x8b11 SIOCGIWSPY    */
        NULL,                           /* 0x8b12 SIOCSIWTHRSPY */
        NULL,                           /* 0x8b13 SIOCGIWTHRSPY */

        owl_wext_set_ap,                /* 0x8b14 SIOCSIWAP     */
        owl_wext_get_ap,                /* 0x8b15 SIOCGIWAP     */
        owl_wext_set_mlme,              /* 0x8b16 SIOCSIWMLME   */
        NULL,                           /* 0x8b17 SIOCGIWAPLIST */
        owl_wext_set_scan,              /* 0x8b18 SIOCSIWSCAN   */
        owl_wext_get_scan,              /* 0x8b19 SIOCGIWSCAN   */

        owl_wext_set_essid,             /* 0x8b1a SIOCSIWESSID  */
        owl_wext_get_essid,             /* 0x8b1b SIOCGIWESSID  */
        NULL,                           /* 0x8b1c SIOCSIWNICKN  */
        NULL,                           /* 0x8b1d SIOCGIWNICKN  */

        NULL,                           /* 0x8b1e */
        NULL,                           /* 0x8b1f */
        NULL,                           /* 0x8b20 SIOCSIWRATE   */
        owl_wext_get_rate,              /* 0x8b21 SIOCGIWRATE   */
        NULL,                           /* 0x8b22 SIOCSIWRTS    */
        NULL,                           /* 0x8b23 SIOCGIWRTS    */
        NULL,                           /* 0x8b24 SIOCSIWFRAG   */
        NULL,                           /* 0x8b25 SIOCGIWFRAG   */
        NULL,                           /* 0x8b26 SIOCSIWTXPOW  */
        NULL,                           /* 0x8b27 SIOCGIWTXPOW  */
        NULL,                           /* 0x8b28 SIOCSIWRETRY  */
        NULL,                           /* 0x8b29 SIOCGIWRETRY  */

        owl_wext_set_encode,            /* 0x8b2a SIOCSIWENCODE */
        owl_wext_get_encode,            /* 0x8b2b SIOCGIWENCODE */

        owl_wext_set_power,             /* 0x8b2c SIOCSIWPOWER  */
        owl_wext_get_power,             /* 0x8b2d SIOCGIWPOWER  */

        NULL,                           /* 0x8b2e */
        NULL,                           /* 0x8b2f */
        owl_wext_set_genie,             /* 0x8b30 SIOCSIWGENIE  */
        NULL,                           /* 0x8b31 SIOCGIWGENIE  */

        owl_wext_set_auth,              /* 0x8b32 SIOCSIWAUTH   */
        NULL,                           /* 0x8b33 SIOCGIWAUTH   */

        owl_wext_set_encode_ext,        /* 0x8b34 SIOCSIWENCODEEXT   */
        NULL,                           /* 0x8b35 SIOCGIWENCODEEXT   */

        NULL,                           /* 0x8b36 SIOCSIWPMKSA  */
};


const struct iw_handler_def owl_wireless_handlers = {
        .standard = owl_iw_handler,
        .num_standard = ARRAY_SIZE(owl_iw_handler),
        
#ifdef CONFIG_WEXT_PRIV
        .num_private = 0,
        .num_private_args = 0,
        .private = NULL,
        .private_args = NULL,
#endif

        .get_wireless_stats = owl_wext_get_wireless_stats,
};
